/*  LINALG.C                                                      */

#include <stdio.h>
#include <stdlib.h>
#include "linalg.h"
#include "error.h"

void dok(unit *p, int n) {
  register unit u=1;
  while (n--) {
    if (u & *p) {
      printf("1");
    } else printf("0");
    u <<=1;
    if (!u) { u=1; p++; };
  };
  printf("\n");
}

void dokmat(matrix *m) {
  unsigned k = m->num_r; unit *p=m->dat[0];
  while (k--) {
    dok(p,m->num_c); p+=UNITSPERROW;
  };
}

void dokminimat(minimatrix *m) {
  unsigned k = m->num_r; unit *p=m->dat[0];
  while (k--) { dok(p,m->num_c); p+=UNITSPERROW; };
}


__inline__ void add_row(unit *p, unit *q, unsigned k) {
  register unit *s = p, *t = q;
  do { *s++ ^= *t++; } while (k--);
}

__inline__ void copy_row(unit *p, unit *q, unsigned k) {
  register unit *s = p, *t = q;
  do { *s++ = *t++; } while (k--);
}

__inline__ void clear_row(unit *p, unsigned k) {
  register unit *q,*s; s = q = p;
  do { *s ^= *s; s++; } while (k--);
  p = q;
}

void clear_matrix(matrix *m, unsigned z, unsigned s) {
  unit *p=m->dat[0];
  m->num_r = z; m->num_c = s; m->units_per_row = (s >>= UNSIZE) + 1; s++;
  while (z--) { clear_row(p,s); p += UNITSPERROW; };
}

int matrix_is_zero(matrix *m) {
  unit *p=m->dat[0];
  int z = m->num_r;
  for (;z--;p+=UNITSPERROW) {
    unit *q=p; int c = m->units_per_row;
    while (c--) if (*q++) return 0;
  }
  return 1;
}

__inline__ int set_values(matrix *m, unsigned z, unsigned s) {
  m->num_r = z; m->num_c = s; m->units_per_row = (s >> UNSIZE)+1;
  /*   clear_matrix(m,z,s);  */
  return m->units_per_row;
}

void unit_matrix(matrix *m, unsigned z) {
  unit u, *l=m->dat[0];
  clear_matrix(m,z,z);
  u = 1;
  while (z--) {
    *l = u; if (!(u <<= 1)) { u = 1; l++; };
    l += UNITSPERROW;
  };
}

void matrix_copy(matrix *m, matrix *n) { /* m := n */
  register unsigned k=(m->num_r = n->num_r);
  register unsigned s,ss=(m->units_per_row = n->units_per_row);
  register unit *mp=m->dat[0],*np=n->dat[0];
  m->num_c = n->num_c;
  while (k--)
    { s=ss; copy_row(mp,np,s); mp += UNITSPERROW; np += UNITSPERROW; };
}

matrix dummy;

#define mask(x) (x ^ (x & ((x)-1)))

void kernel_basis(matrix *m, matrix *n) {  /* n := basis of kernel of m */
  unsigned k,t,o,oo,rr;
  unit *xp,*yp,*zp,*dp=dummy.dat[0], *mp=m->dat[0], *np=n->dat[0], u;
  n->num_r = 0; n->num_c = m->num_r;
  unit_matrix(&dummy,m->num_r);
  oo = n->units_per_row = dummy.units_per_row;
  rr = m->units_per_row;
  for (k=0;k<m->num_r;k++, dp += UNITSPERROW, mp += UNITSPERROW) {
    xp = mp; o = rr;
    while ((u = mask(*xp)) == 0 && o--) xp++;
    if (u) {
      yp = dp; zp = mp;
      for (t=k+1; t<m->num_r; t++) {
	xp += UNITSPERROW; yp += UNITSPERROW; zp += UNITSPERROW;
	if (*xp & u) {
	  add_row(yp,dp,oo); add_row(zp,mp,rr);
	};
      };
    }
    else {
      copy_row(np,dp,oo);
      n->num_r++; np += UNITSPERROW;
    };
  };
}


void quotient_basis(matrix *n, matrix *m, minimatrix *r) { 
  /* r := basis of (Im(m)+Im(n))/Im(n) */
  unsigned k,l,o,oo; unit *mp,*np,*xp,*yp,*zp,u;
  if (m->num_c != n->num_c) { printf("quotient_basis ERROR!\n"); exit(1); };
  oo = (r->num_c = m->num_c) >> UNSIZE; r->num_r = 0; 
  r->units_per_row = n->units_per_row;
  mp = m->dat[0]; np = n->dat[0];
  for (k=0; k<m->num_r; k++, mp += UNITSPERROW) {
    xp = mp; yp = np; o = oo;
    while ((u = mask(*xp)) == 0 && o--) { xp++; yp++; };
    if (u) {
      zp = mp;
      for (l=k+1; l<m->num_r; l++) {
	xp += UNITSPERROW; zp += UNITSPERROW;
	if (*xp & u) add_row(zp,mp,oo);
      };
      zp = np;
      for (l=0; l<n->num_r; l++) {
	if (*yp & u) add_row(zp,mp,oo);
	yp += UNITSPERROW; zp += UNITSPERROW;
      };
    };
  };
  np = n->dat[0];
  for (k=0; k<n->num_r; k++, np += UNITSPERROW) {
    xp = np; o = oo;
    while ((u = mask(*xp)) == 0 && o--) xp++;
    if (u != 0) {
      copy_row(r->dat[r->num_r++],np,oo);
      zp = np;
      for (l=k+1; l<n->num_r; l++) {
	xp += UNITSPERROW; zp += UNITSPERROW;
	if (*xp & u) add_row(zp,np,oo);
      };
    } else {
    };
  };
}

/* rearranged lift 11/11/98 to start at the bottom right corner */
int lift(minimatrix *mini, matrix *mat, minimatrix *res) {
  int oores,todo,p,k,l;
  if (mat->num_c != mini->num_c) {
    printf("mini = %d, mat = %d\n",mini->num_c,mat->num_c);
    error("inconsistent matrices");
  }
  unit_matrix(&dummy,mat->num_r);
  clear_matrix((matrix *) res, mini->num_r, mat->num_r);
  oores = set_values((matrix *) res, mini->num_r, mat->num_r);
  todo = mini->num_r; /* number of vectors to lift */
  /* exclude zero vectors: */
  for (k=mini->num_r;k--;) {
    for (l=0;(0 == mini->dat[k][l]) && (l<mini->units_per_row);l++) ;
    if (l>=mini->units_per_row) todo--;
  };
  /*
    printf("todo=%d,",todo);
    printf("mini =\n"); dokminimat(mini);
    printf("mat =\n"); dokmat(mat);
    printf("dummy =\n"); dokmat(&dummy);
    printf("res =\n"); dokminimat(res);
  */
  /* make mat orthogonal, thereby reducing mini */

  for (k=mat->num_r;k--;) {
    unit umat=0; int hmat; /* (hmat,umat) describes row in mat */
    /* find first 1 in mat->dat[k] */
    for (l=mat->units_per_row;(l--) && (0 == (umat = mask(mat->dat[k][l])));) ;
    if (0 != umat) { /* found a 1 */
      hmat = l;
      for (p=0;p<mat->num_r;p++)
	if ((p != k) && (mat->dat[p][hmat] & umat) != 0) {
	  add_row(mat->dat[p],mat->dat[k],mat->units_per_row);
	  add_row(dummy.dat[p],dummy.dat[k],dummy.units_per_row);
	}
      for (p=0;p<mini->num_r;p++)
	if ((mini->dat[p][hmat] & umat) != 0) {
	  add_row(mini->dat[p],mat->dat[k],mat->units_per_row);
	  add_row(res->dat[p],dummy.dat[k],dummy.units_per_row);
	  for (l=0;(0 == mini->dat[p][l]) && (l<mini->units_per_row);l++) ;
	  if (l >= mini->units_per_row) todo--;
	}
    }
    if (!todo) break;
  }
  
  if (todo) {
    printf("mini =\n"); dokminimat(mini);
    printf("mat =\n"); dokmat(mat);
    printf("dummy =\n"); dokmat(&dummy);
    printf("res =\n"); dokminimat(res);
    if (todo) return 0;
    /* error("Lift(s) not found");  */
  }
  
  return 1;
}
