/*  options.c  */

#include "options.h"
#include <stdio.h>

int arg_s, arg_t;

int op_per=1;      /* print percentage during computation ?       */
int op_recomp=0;   /* recompute known bidegrees ?                 */
int op_test=0;     /* testmodus -- abort if new generators needed */
int op_quiet=0;    /* minimal output                              */  
int op_pc=0;       /* specialities for the parallel computation   */

int op_alg=-1;     /* specifies the algebra A(n) to work with     */

char logfileextension[30];

char *def(int n) {
  return n ? "yes" : " no";
} 

void usage(void) {
  printf("usage:\n");
  printf(" resolve s tmax <options>\n");
  printf("options are:\n");
  printf(" +/-per    -> percentage information during computation    (def: %s)\n",def(op_per));
  printf(" +/-recomp -> recompute known bidegrees                    (def: %s)\n",def(op_recomp));
  printf(" +/-test   -> test modus, abort if new gens needed         (def: %s)\n",def(op_test));
  printf(" +/-quiet  -> only minimal output                          (def: %s)\n",def(op_quiet));
  printf(" +/-pc     -> specialities for the parallel computation    (def: %s)\n",def(op_pc));
  printf(" logext:<name>   -> logfile extension (default: none)\n");
  printf(" alg:<value>     -> specifies A(<value>)\n");
  exit(1);
}

int badopt = 0;

void bad_option(char *p) {
  printf("invalid option: %s\n",p); badopt = 1;
}

int streq(char *a, char *b, int n) {
  while (n--) if (*a++ != *b++) return 0;
  return 1;
}

void handle_option(char *op) {
  if ((*op == '+') || (*op == '-')) {
    int bool = (*op++ == '+');
    if (streq(op,"per",3)) { op_per = bool; return; }
    if (streq(op,"recomp",3)) { op_recomp = bool; return; }
    if (streq(op,"test",4)) { op_test = bool; return; }
    if (streq(op,"quiet",5)) { op_quiet = bool; return; }
    if (streq(op,"pc",2)) { op_pc = bool; return; }
    op--; bad_option(op); return;
  }
  if (1 == sscanf(op,"logext:%s",logfileextension)) {
    if (0 == *logfileextension) { 
      printf("empty logfile extension (file \"log.\" not allowed)\n");
      badopt = 1;
    }
    return;
  }
  if (1 == sscanf(op,"alg:%d",&op_alg)) {
    if ((op_alg < 0) || (op_alg > 7)) {
      printf("subalgebra A(%d) not accepted\n",op_alg);
      badopt = 1;
    }
    return;
  }
  bad_option(op);
}

void parse_options(int argc, char **argv) {
  int n;

  logfileextension[0] = 0;

  if (argc < 3) usage();

  if (1 != sscanf(argv[1],"%d",&arg_s)) {
    printf("error: 1st argument (%s) is not a number\n",argv[1]);
    badopt=1;
  }

  if (1 != sscanf(argv[2],"%d",&arg_t)) {
    printf("error: 2nd argument (%s) is not a number\n",argv[2]);
    badopt=1;
  }

  for (n=3;n<argc;n++) handle_option(argv[n]);

  if (badopt) usage();
}

/* 
int main(int argc, char **argv) {
  parse_options(argc,argv);
  printf("** OPTIONS OK, arg_s=%d, arg_t=%d, logfileextension = %s\n",
         arg_s,arg_t,logfileextension);
  usage();
  return 0;
}
*/
