/* stop.c   -   Steenrod operations and related things */

#include "stop.h"
#include "error.h"

int MAXDEGREE = 0;

int degrees[12] = {1,3,7,15,31,63,127,255,511,1023,2047,4095};

void initialize_stops(void) {
   MAXDEGREE = (1 << NALG) - 1;
   printf("stop.c 11/23/98, NALG = %d, MAXDEGREE = %d\n",NALG,MAXDEGREE);
}

void clear_stop(stop *s) {
   int n;
   for (n=NALG;n--;) s->dat[n] = 0;
}

void copy_stop(stop *s, stop *t) {
   int n;
   for (n=NALG;n--;) s->dat[n] = t->dat[n];
   s->id = t->id;
}

void double_stop(stop *s) {
  int n; 
  for (n=NALG;n--;) s->dat[n] <<= 1 ;
}

int halve_stop(stop *s) {
  int n,r=0;
  for (n=NALG;n--;) { r |= s->dat[n]; s->dat[n] >>= 1; }
  return !(r & 1);
}

void invert_stop(stop *s) {
  int n;
  for (n=NALG;n--;) s->dat[n] = ~(s->dat[n]);
}

int degree_stop(stop *s, int *deglist) {
   int res=0,n=0;
   while (n<NALG) { res += s->dat[n] * deglist[n]; n++; };
   return res;
}

int stops_are_equal(stop *a,stop *b) {
  int n;
  if (a->id != b->id) return 0;
  for (n=0;n<NALG;n++) if (a->dat[n] != b->dat[n]) return 0;
  return 1;
}

#define ENDDIFF -1

#define CST3  0
#define CST5  1
#define CST7  2
#define CST10 3

void write_enddiff(FILE *fp) {
   stop aux;
   aux.id = ENDDIFF;
   write_stop(fp,&aux);
}

void fprintf_stop(FILE *fp, stop *s) {
   int n=NALG,m; char c='(';
   if (s->id == ENDDIFF) {
      fprintf(fp,"<end>\n");
      return;
   }
   while (n-- && (s->dat[n] == 0)) ;
   if (s->dat[n] != 0) {
      fprintf(fp,"Sq");
      for (m=0;m<=n;m++,c=',') fprintf(fp,"%c%d",c,s->dat[m]);
      fprintf(fp,")");
   }
   fprintf(fp,"[%d]",s->id);
}

void print_stop(stop *s) { fprintf_stop(stdout,s); }

typedef int bigstop[12];

void clearbigstop(bigstop x) {
  int n;
  for (n=12;n--;) x[n] = 0;
}

void makebigstop_from_stop(bigstop x, stop *s) {
  int n;
  clearbigstop(x);
  for (n=NALG;n--;) x[n] = s->dat[n];
}

void makestop_from_bigstop(stop *s, bigstop x) {
  int n;
  for (n=NALG;n--;) s->dat[n] = x[n];
}

int stop_fits_n(stop *s, int n) {
  /* s->dat[k] is allowed to have n-k bits */
  int i,j;
  j = (1 << n) - 1; /* this has n ones */
  for (i=0;i<NALG;i++,j>>=1) {
    if (s->dat[i] != (j & s->dat[i])) return 0;
  }
  return 1;
}

uint64 wrst;
uint16 wrid;

void pack(stop *t, int n) {
  int sh=1; bigstop s;
  makebigstop_from_stop(s,t);
  wrst=0;
  for (;n--;sh++) { wrst <<= sh; wrst |= s[n]; }
}

void unpack(stop *s, int n) {
  uint64 msk;
  int shft  = n; int k;
  bigstop x;
  clearbigstop(x);
  msk = 1; msk <<= n; msk--; /* msk has n ones */
  for (k=0;n--;msk >>= 1,shft--,k++) {
    x[k] = wrst & msk;
    wrst >>= shft;
  }
  makestop_from_bigstop(s,x);
}

void write_stop_3(FILE *fp, stop *t) {
  uint8 x; uint64 tst;
  pack(t,3); wrid = CST3; wrid <<= 13; wrid |= ((1 << 13) - 1) & t->id;
  x = wrst; tst = x;
  if (tst != wrst) error("write_stop_3(...) err#1");
  if (1 != fwrite(&wrid,sizeof(int16),1,fp))
    error("write_stop_3(...) err#2");  
  if (1 != fwrite(&x,sizeof(int8),1,fp)) 
    error("write_stop_3(...) err#3");  
}

void write_stop_5(FILE *fp, stop *t) {
  uint16 x; uint64 tst;
  pack(t,5); wrid = CST5; wrid <<= 13; wrid |= ((1 << 13) - 1) & t->id;
  x = wrst; tst = x;
  if (tst != wrst) error("write_stop_5(...) err#1");
  if (1 != fwrite(&wrid,sizeof(int16),1,fp))
    error("write_stop_5(...) err#2");  
  if (1 != fwrite(&x,sizeof(int16),1,fp)) 
    error("write_stop_5(...) err#3");  
}

void write_stop_7(FILE *fp, stop *t) {
  uint32 x; uint64 tst;
  pack(t,7); wrid = CST7; wrid <<= 13; wrid |= ((1 << 13) - 1) & t->id;
  x = wrst; tst = x;
  if (tst != wrst) error("write_stop_7(...) err#1");
  if (1 != fwrite(&wrid,sizeof(int16),1,fp))
    error("write_stop_7(...) err#2");  
  if (1 != fwrite(&x,sizeof(int32),1,fp)) 
    error("write_stop_7(...) err#3");  
}

void write_stop_10(FILE *fp, stop *t) {
  uint64 x; uint64 tst;
  pack(t,10); wrid = CST10; wrid <<= 13; wrid |= ((1 << 13) - 1) & t->id;
  x = wrst; tst = x;
  if (tst != wrst) error("write_stop_10(...) err#1");
  if (1 != fwrite(&wrid,sizeof(int16),1,fp))
    error("write_stop_10(...) err#2");  
  if (1 != fwrite(&x,sizeof(int64),1,fp)) 
    error("write_stop_10(...) err#3");  
}

void write_stop(FILE *fp, stop *s) {
  stop z;
  z.id = ENDDIFF;
  if (s->id == z.id) { write_stop_10(fp,s); return; }
  if (stop_fits_n(s,3))  { write_stop_3(fp,s);  return; }
  if (stop_fits_n(s,5))  { write_stop_5(fp,s);  return; }
  if (stop_fits_n(s,7))  { write_stop_7(fp,s);  return; }
  if (stop_fits_n(s,10)) { write_stop_10(fp,s); return; }
  printf("write_stop(...,");
  fprintf_stop(stdout,s);
  printf(")\n");
  error("write_stop(...): stop too big");
}

int fscanf_stop(FILE *fp, stop *s) {
   int n; int aux;
   if (1 == fscanf(fp," *%n",&n)) return 0;
   clear_stop(s);
   if (1 == fscanf(fp," Sq(%d",&aux)) { 
      s->dat[0] = aux;
      n = 1;
      while (1 == fscanf(fp,",%d",&aux)) { s->dat[n] = aux; n++; }
      if (n>NALG) error("fscanf_stop: NALG too small");
      if (1 != fscanf(fp,")[%hd]",&(s->id))) error("fscanf_stop: [id] expected");
      return 1;
   }
   return 0;
}

int read_stop(FILE *fp, stop *s) {
  uint8 a8; uint16 a16; uint32 a32; uint64 a64;
  uint16 msk;
  msk = 1; msk <<= 13; msk--; /* msk has 13 bits */
  if (1 != fread(&wrid,sizeof(int16),1,fp))
    error("read_stop(...) err#1");
  s->id = wrid & msk;
  wrid >>= 13;
  /*printf("{%d}",wrid);*/
  switch (wrid) {
  case CST3:
    if (1 != fread(&a8,sizeof(int8),1,fp))
      error("read_stop(...) err#2 (int8)");
    wrst = a8;
    unpack(s,3);
    break;
  case CST5:
    if (1 != fread(&a16,sizeof(int16),1,fp))
      error("read_stop(...) err#2 (int16)");
    wrst = a16;
    unpack(s,5);
    break;
  case CST7:
    if (1 != fread(&a32,sizeof(int32),1,fp))
      error("read_stop(...) err#2 (int32)");
    wrst = a32;
    unpack(s,7);
    break;
  case CST10:
    if (1 != fread(&a64,sizeof(int64),1,fp))
      error("read_stop(...) err#2 (int64)");
    wrst = a64;
    unpack(s,10);
    break;
  default :
    error("read_stop(...) err#3");
  }
  a16 = ENDDIFF;
  if ((a16 & msk) == (s->id & msk)) return 0;
  /*fprintf_stop(stdout,s);*/
  return 1;
}
